#include <stdlib.h>
#include <setjmp.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>

long pageSize;
sigjmp_buf tryReadEnv;

void SIGSEGV_Handler (int signum);

void Init (void)
{
// VAR
	void (*res)(int);
// BEGIN
	pageSize = sysconf(_SC_PAGESIZE); assert(pageSize > 0);
	res = signal(SIGSEGV, SIGSEGV_Handler); assert(res != SIG_ERR);
}

int IsReadable (int from, int to)
{
// VAR
	int res;
// BEGIN
	from -= from % pageSize;
	//to += pageSize - to % pageSize;
	res = madvise((void *)from, to - from, MADV_NORMAL); assert(res != -1 || errno != EINVAL);
	return res == 0;
}

void SIGSEGV_Handler (int signum)
{
	siglongjmp(tryReadEnv, 1);
}

void TryRead (int from, int to, int *ok)
{
// VAR
	char *p, ch; int len, res;
// BEGIN
	res = sigsetjmp(tryReadEnv, 1);
	if (res == 0)
	{
		p = (char *)from; len = to - from;
		while (len > 0) { ch = *p; p++; len--; }
		*ok = 1;
	}
	else
	{
		*ok = 0;
	}
}

void TestRead (int p1, int p2, int r[4])
{
// VAR
	int i; int to, from;
	int readable, readed;
// BEGIN
	assert(p1 <= p2);
	for (i = 0; i <= 3; ++i) r[i] = 0;
	for (i = p1; i <= p2; ++i)
	{
		printf("#%i\r", i); fflush(stdout);
		from = i * pageSize; to = from + pageSize;
		readable = IsReadable(from, to);
		TryRead(from, to, &readed);
		//printf("%08X - %08X: %i %i\n", from, to, readable, readed);
		++r[2 * readable + readed];
	}
}

void ShowSConfValue (int n, const char *pt)
{
	int res;
	res = sysconf(n);
	printf("%s: ", pt);
	if (res != -1) printf("%i\n", res); else printf("value unavailable\n");
}

void ShowInfo ()
{
	ShowSConfValue(_SC_PAGESIZE, "pageSize");
	ShowSConfValue(_SC_PHYS_PAGES, "Number of pages of physycal memory");
	ShowSConfValue(_SC_AVPHYS_PAGES, "Number of currently available pages of physycal memory");
}

void StrToInt (const char *str, int *x, int *ok)
{
	int res;
	res = sscanf(str, "%i", x);
	*ok = res == 1;
}

void ReadArgs (int argc, char **argv, int *ok, int *a1, int *a2, int *a3)
{
	int res;
	*ok = (3 <= argc) && (argc <= 4);
	if (*ok) StrToInt(argv[1], a1, ok);
	if (*ok) StrToInt(argv[2], a2, ok);
	if (*ok && argc == 4)
	{
		StrToInt(argv[3], a3, &res);
		if (!res) a3 = 0;
	}
}

int main (int argc, char *argv[])
{
// VAR
	int p1, p2, bytes, ok;
	int r[4];
	void *mem = NULL;
// BEGIN
	Init();
	ReadArgs(argc, argv, &ok, &p1, &p2, &bytes);
	if (ok)
	{
		assert(p1 <= p2);
		ShowInfo();
		if (bytes > 0)
		{
			printf("Allocating %i bytes...", bytes);
			mem = malloc(bytes);
			if (mem != NULL) printf("ok\n"); else printf("error\n");
		}
		printf("Testing %i pages from %i to %i\n", p2 - p1 + 1, p1, p2);
		TestRead(p1, p2, r);
		printf("~readable & ~readed: %i\n", r[0]);
		printf("~readable & readed: %i\n", r[1]);
		printf("readable & ~readed: %i\n", r[2]);
		printf("readable & readed: %i\n", r[3]);
		if (mem) free(mem);
	}
	else
	{
		printf("is_readable fromPage toPage [allocBytes]\n");
		ShowInfo();
	}
	return 0;
}
